#!/usr/bin/env python
#
# Copyright (c) Members of the EGEE Collaboration. 2004.
# See http://www.eu-egee.org/partners/ for details on the copyright
# holders.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# GLite Data Transfer Load Generator : file names generation script
#
# Authors: Paolo Tedesco <paolo.tedesco@cern.ch>
# Version info: $Id: generate_load.py,v 1.3 2008/06/05 15:35:28 ptedesco Exp $
# Release: $Name:  $
#
import commands
import os
import random
import re
import sys
import time
import urllib

#-------------------------------------------------------------------------------
# Stub TS errors generation
#-------------------------------------------------------------------------------
#class Status:
#    """Collection of transfer error strings."""
#    RECEIVED = "S_RECEIVED"
#    PREPARING = "S_PREPARING"
#    READY = "S_READY"
#    SUBMITTED = "S_SUBMITTED"
#    PROCESSING = "S_PROCESSING"
#    TRANSFERRED = "S_TRANSFERRED"
#    COMPLETED = "S_COMPLETED"
#    FAILED = "S_FAILED"
#    ABORTED = "S_ABORTED"

class C:
    """Collection of transfer error category strings."""
    TRANSFER = "TRANSFER"
    SECURITY = "SECURITY"
    NETWORK = "NETWORK"
    CONFIGURATION = "CONFIGURATION"
    CONSISTENCY = "CONSISTENCY"
    CATALOG = "CATALOG"
    USER = "USER"
    FILE_EXISTS = "FILE_EXISTS"
    INVALID_PATH = "INVALID_PATH"
    NO_SPACE_LEFT = "NO_SPACE_LEFT"
    INVALID_SIZE = "INVALID_SIZE"
    REQUEST_FAILURE = "REQUEST_FAILURE"
    CONNECTION = "CONNECTION"
    HTTP_TIMEOUT = "HTTP_TIMEOUT"
    REQUEST_TIMEOUT = "REQUEST_TIMEOUT"
    TRANSFER_TIMEOUT = "TRANSFER_TIMEOUT"
    INTERNAL_ERROR = "INTERNAL_ERROR"
    ABORTED = "ABORTED"
    GENERAL_FAILURE = "GENERAL_FAILURE"
    LOST = "LOST"
    STORAGE_INTERNAL_ERROR = "STORAGE_INTERNAL_ERROR"
    GRIDFTP = "GRIDFTP"
    MARKERS_TIMEOUT = "MARKERS_TIMEOUT"
    SRM_COPY = "SRM_COPY"
    CATALOG_INTERNAL_ERROR = "CATALOG_INTERNAL_ERROR"

class P:
    """Collection of transfer error phase strings."""
    TX_PREPARATION = "P_TX_PREPARATION"
    TX_FINALIZATION = "P_TX_FINALIZATION"
    TRANSFER = "P_TRANSFER"
    TX_SERVICE = "P_TX_SERVICE"

class S:
    """Collection of transfer error scope strings."""
    SOURCE = "S_SOURCE"
    DESTINATION = "S_DESTINATION"
    TRANSFER = "S_TRANSFER"
    AGENT = "S_AGENT"

class E:
    """An error."""
    
    def __init__(self,occurrencies,c,p,s,m=None):
        """Constructor.
            occurrencies: the number of occurrencies for the error.
            c: error category
            p: error phase
            s: error scope
            m: error message (optional)
        """
        self.occurrencies = occurrencies
        self.name = "fail-S_PROCESSING-c-%s-p-%s-s-%s" % (c,s,p)
        if not m is None:
            self.name = self.name + "-m-" + urllib.quote(m)

class StubErrorGenerator:
    """Creates file names that cause errors in the stub transfer service."""
    def __init__(self):
        # total error occurrencies
        self.totalOccurrencies = 0
        # errors list
        self.errors = []
        self._add(1253, C.REQUEST_TIMEOUT,          S.SOURCE,       P.TX_PREPARATION,   "failed to prepare source file in 180 seconds")
        self._add(1148, C.REQUEST_TIMEOUT,          S.DESTINATION,  P.TX_PREPARATION,   "failed to prepare destination file in 180 seconds")
        self._add(945,  C.TRANSFER_TIMEOUT,         S.TRANSFER,     P.TRANSFER,         "gridftp_copy_wait: Connection timed out")
        self._add(928,  C.GRIDFTP,                  S.TRANSFER,     P.TRANSFER,         "the server sent an error response: 425 425 Can't open data connection. timed out() failed.")
        self._add(724,  C.NO_SPACE_LEFT,            S.TRANSFER,     P.TRANSFER,         "an end-of-file was reached")
        self._add(648,  C.CONNECTION,               S.DESTINATION,  P.TX_PREPARATION,   "failed to contact on remote SRM [srm]. Givin' up after 3 tries")
        self._add(583,  C.STORAGE_INTERNAL_ERROR,   S.DESTINATION,  P.TX_PREPARATION,   "CastorStagerInterface.c:2507 Device or resource busy (errno=0, serrno=0)")
        self._add(560,  C.GRIDFTP,                  S.TRANSFER,     P.TRANSFER,         "globus_gass_copy_register_url_to_url: Connection timed out")
        self._add(527,  C.GRIDFTP,                  S.TRANSFER,     P.TX_FINALIZATION,  "empty file size returned")
        self._add(506,  C.INVALID_PATH,             S.SOURCE,       P.TX_PREPARATION,   "specified file(s) does not exist")
        self._add(450,  C.SECURITY,                 S.TRANSFER,     P.TRANSFER,         "the server sent an error response: 553 553 [addres]: Permission denied")
        self._add(428,  C.GRIDFTP,                  S.SOURCE,       P.TRANSFER,         "globus_ftp_client_size: Connection timed out")
        self._add(410,  C.STORAGE_INTERNAL_ERROR,   S.SOURCE,       P.TX_PREPARATION,   "CastorStagerInterface.c:2162 Required tape segments are not all accessible (errno=0, serrno=0)")
        self._add(408,  C.GRIDFTP,                  S.TRANSFER,     P.TRANSFER,         "the server sent an error response: 426 426 Data connection. data write() failed: Handle not in the proper state")
        self._add(405,  C.STORAGE_INTERNAL_ERROR,   S.DESTINATION,  P.TX_PREPARATION,   "destination file failed on the SRM with error [SRM FAILURE]")
        self._add(401,  C.INVALID_SIZE,             S.SOURCE,       P.TX_PREPARATION,   "cannot continue since no size has been returned after PrepareToGet or SrmStat")
        self._add(392,  C.FILE_EXISTS,              S.DESTINATION,  P.TX_PREPARATION,   "RequestFileStatus#[id] failed with error:[ [DATE] state Failed : GetStorageInfoFailed : file exists, cannot write")
        self._add(371,  C.INVALID_PATH,             S.SOURCE,       P.TX_PREPARATION,   "Failed on SRM get: Failed SRM get on [addres] call. Error is RequestFileStatus#-[] failed with error:[ at Wed Feb 21 12:18:44 CET 2007 state Failed : file not found : path [path] not found")
        self._add(370,  C.STORAGE_INTERNAL_ERROR,   S.TRANSFER,     P.TRANSFER,         "451 Local resource failure: malloc: Cannot allocate memory")
        self._add(304,  C.GRIDFTP,                  S.TRANSFER,     P.TRANSFER,         "the server sent an error response: 421 421 Timeout (900 seconds): closing control connection.")
        self._add(265,  C.GRIDFTP,                  S.TRANSFER,     P.TRANSFER,         "the server sent an error response: 451 451 rfio read failure")
        self._add(260,  C.GRIDFTP,                  S.TRANSFER,     P.TRANSFER,         "the server sent an error response: 550 550 [addres] not a plain file.")
        self._add(249,  C.GRIDFTP,                  S.TRANSFER,     P.TRANSFER,         "the server sent an error response: 550 550 [addres] Timed out.")
        self._add(248,  C.GRIDFTP,                  S.TRANSFER,     P.TRANSFER,         "a system call failed (Connection refused)")
        self._add(234,  C.STORAGE_INTERNAL_ERROR,   S.SOURCE,       P.TX_PREPARATION,   "Failed on SRM get: Failed SRM get on [addres] ; id=[id] call. Error is CastorStagerInterface.c:2145 Internal error (errno=0, serrno=1015)")
    
    def generate(self):
        irnd = random.randint(1,self.totalOccurrencies)
        icur = 0
        for e in self.errors:
            icur = icur + e.occurrencies
            if icur >= irnd:
                return e.name
        return "xxx"

    def _add(self,o,c,p,s,m=None):
        self.totalOccurrencies = self.totalOccurrencies + o
        self.errors.append(E(o,c,p,s,m))
                

#-------------------------------------------------------------------------------
#                             Pattern substitution
#-------------------------------------------------------------------------------
class PatternReplacer:
    """Replaces patterns in expressions:
        ${name} is expanded to environment variable 'name'
        %{name} or %{name (args)} is expanded to the result of the call to 
            'self._name(args)'.
    """ 

    def __init__(self):
        self.__date = time.strftime("%Y%m%d") 
        self.__time = time.strftime("%H%M%S")
        self.__regexp = re.compile("""
            \${(?P<env>\w+)}
            |
            %{
                \s*(?P<fun>\w+)
                \s*(?P<args>\(.*\))?\s*
            }
            """,re.VERBOSE) 
        self.__stubErrorGenerator = StubErrorGenerator()

    def replace(self,pattern):
        """Replace patterns.""" 
        return self.__regexp.sub(self.__replace,pattern)     
                    
    def _date(self):
        """Current date.""" 
        return self.__date
    
    def _time(self):
        """Current time.""" 
        return self.__time
    
    def _random(self,min = 1,max = 100,format = "%03d"):
        """A random number between min and max, formatted using given format 
            string.""" 
        imin = int(min)
        imax = int(max)
        irnd = random.randint(imin,imax)
        return format % (irnd)
    
    def _guid(self):
        """A guid.""" 
        return commands.getoutput("uuidgen")
	
    def _error(self):
        """An error for the stub transfer service."""
        return self.__stubErrorGenerator.generate()

    def __replace(self,match):
        """Actual replacement function.""" 
        if match.group("env"):
            value = match.group("env")
            return os.environ[value]
        elif match.group("fun"):
            value = match.group("fun")
            args = match.group("args")
            if args is None:
                to_exec = "ret = self._%s()" % (value)
            else: 
                to_exec = "ret = self._%s %s" % (value,args)
            exec (to_exec)
            return ret 
        else:
            error("Error in patterns substitution")
            return "" 

if __name__ == "__main__":

    #---------------------------------------------------------------------------
    #                               main
    #---------------------------------------------------------------------------
    def main (argv = None):
        if argv is None:
            argv = sys.argv
        if 1 == len(argv) or argv[1] in ["-h","--help"]:
            print "Usage: " + argv[0] + " <pattern> (<pattern>...)"
            return 1
        try:
            pr=PatternReplacer()
            for pattern in argv[1:]:
                print pr.replace(pattern)
        except Exception,exc:
            print "error generating file names : " + str(exc)
            return 1

        return 0

    # run main
    sys.exit (main())

else:

    import glite.data.agents.utils
    import glite.load_generator

    #---------------------------------------------------------------------------
    #                                   Logging
    #---------------------------------------------------------------------------

    CATEGORY = "generate_load"

    # debug
    def debug(msg):
        """Log a debug message.
            msg: the message"""
        glite.data.agents.utils.LogDebug(CATEGORY,msg)
        
    # info
    def info(msg):
        """Log an informational message.
            msg: the message"""
        glite.data.agents.utils.LogInfo(CATEGORY,msg)
        
    # warn
    def warn(msg):
        """Log a warning message.
            msg: the message"""
        glite.data.agents.utils.LogWarn(CATEGORY,msg)
        
    # error 
    def error(msg):
        """Log an error message.
            msg: the message"""
        glite.data.agents.utils.LogError(CATEGORY,msg)

    #---------------------------------------------------------------------------
    #                               Name generation
    #---------------------------------------------------------------------------

    # combine
    def combine(a,b):
        """Combine two paths."""
        separator = "/"
        if a.endswith(separator):
            a = a.rstrip(separator) 
        if b.startswith(separator):
            b = b.lstrip(separator)
        return a + separator + b
        
    # generate
    def generate(channelName,jobSize,source,destination):
        """Generate couples of source/destination file names.
            channelName: the name of the channel
            jobSize: the number of files per job 
            source: source endpoint data
            destination: destination endpoint data 
        """ 

        debug ("generating %d file names for channel %s" % (jobSize,channelName)) 
        debug ("source : path = '%s' pattern = '%s'" % (source.path,source.pattern))
        debug ("destination : path = '%s' pattern = '%s'" % (destination.path,destination.pattern))
        
        source_file_pattern = combine(source.path,source.pattern)
        destination_file_pattern = combine(destination.path,destination.pattern)

        debug ("source_file_pattern = %s" % (source_file_pattern)) 
        debug ("destination_file_pattern = %s" % (destination_file_pattern)) 
           
        ret = glite.load_generator.FileDataArray()
        try:
            pr = PatternReplacer()
            for i in range(0,jobSize):
                fd = glite.load_generator.FileData()
                fd.sourceFile = pr.replace(source_file_pattern)
                fd.destinationFile = pr.replace(destination_file_pattern)
                ret.append(fd)
        except Exception, exc:
            error(str(exc))
        return ret    
    
